﻿using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;
using System.Runtime.CompilerServices;

BenchmarkRunner.Run<ExceptionBenchmarks>();

public class ExceptionBenchmarks {

	static int AddEx(int? a, int? b) {
		if (a is null) {
			throw new ArgumentNullException(nameof(a));
		}
		if (b is null) {
			throw new ArgumentNullException(nameof(b));
		}

		return a.Value + b.Value;
	}

	static int Alg1Ex(int? a, int? b, int? c, int? d) {
		return AddEx(a, b) * AddEx(b, c) * AddEx(c, d);
	}

	[MethodImpl(MethodImplOptions.NoInlining)]
	static int Alg1Ex_NoInlining(int? a, int? b, int? c, int? d) {
		return AddEx(a, b) * AddEx(b, c) * AddEx(c, d);
	}

	static int Alg2Ex(int? a, int? b, int? c, int? d) {
		int result;
		try {
			result = AddEx(a, b);
		} catch (ArgumentNullException) {
			result = 1;
		}

		try {
			result *= AddEx(b, c);
		} catch (ArgumentNullException) {
			// Leave result as is
		}

		try {
			result *= AddEx(c, d);
		} catch (ArgumentNullException) {
			// Leave result as is
		}

		return result;
	}

	//

	[Benchmark]
	public int Alg1Ex_NoExThrown() {
		return Alg1Ex(2, 3, 4, 5);
	}

	[Benchmark]
	public int Alg1Ex_NoExThrown_NoInlining() {
		return Alg1Ex_NoInlining(2, 3, 4, 5);
	}

	[Benchmark]
	public int Alg2Ex_NoExThrown() {
		return Alg2Ex(2, 3, 4, 5);
	}

	[Benchmark]
	public int Alg1Ex_ThrowsOne() {
		try {
			return Alg1Ex(null, null, null, null);
		} catch (ArgumentNullException) {
			return 0;
		}
	}

	[Benchmark]
	public int Alg2Ex_ThrowsAll() {
		return Alg2Ex(null, null, null, null);
	}

	//

	static bool TryAdd(int? a, int? b, out int result) {
		if (a is null || b is null) {
			result = 0;
			return false;
		} else {
			result = a.Value + b.Value;
			return true;
		}
	}

	static bool TryAlg1(int? a, int? b, int? c, int? d, out int result) {
		result = 1;
		if (!TryAdd(a, b, out var r1)) {
			return false;
		}
		result = r1;

		if (!TryAdd(b, c, out var r2)) {
			return false;
		}
		result *= r2;
		
		if (!TryAdd(c, d, out var r3)) {
			return false;
		}
		result *= r3;

		return true;
	}

	static bool TryAlg1_Variant2(int? a, int? b, int? c, int? d, out int result) {
		if (TryAdd(a, b, out var r1) && TryAdd(b, c, out var r2) && TryAdd(c, d, out var r3)) {
			result = r1 * r2 * r3;
			return true;
		} else {
			result = 1;
			return false;
		}
	}

	//

	[Benchmark]
	public int TryAlg1_NotNull() {
		if (TryAlg1(2, 3, 4, 5, out var result)) {
			return result;
		} else {
			return 0;
		}
	}

	[Benchmark]
	public int TryAlg1_Variant2_NotNull() {
		if (TryAlg1_Variant2(2, 3, 4, 5, out var result)) {
			return result;
		} else {
			return 0;
		}
	}

	[Benchmark]
	public int TryAlg1_Null() {
		if (TryAlg1(null, null, null, null, out var result)) {
			return result;
		} else {
			return 0;
		}
	}

	[Benchmark]
	public int TryAlg1_Variant2_Null() {
		if (TryAlg1_Variant2(null, null, null, null, out var result)) {
			return result;
		} else {
			return 0;
		}
	}

	//

	static int? AddNullable(int? a, int? b) {
		return a + b;
	}

	static int? Alg1Nullable(int? a, int? b, int? c, int? d) {
		return AddNullable(a, b) * AddNullable(b, c) * AddNullable(c, d);
	}

	[MethodImpl(MethodImplOptions.NoInlining)]
	static int? Alg1Nullable_NoInlining(int? a, int? b, int? c, int? d) {
		return AddNullable(a, b) * AddNullable(b, c) * AddNullable(c, d);
	}

	static int Alg2Nullable(int? a, int? b, int? c, int? d) {
		int result;

		var r1 = AddNullable(a, b);
		if (r1 is not null) {
			result = r1.Value;
		} else {
			result = 1;
		}

		var r2 = AddNullable(b, c);
		if (r2 is not null) {
			result *= r2.Value;
		}

		var r3 = AddNullable(c, d);
		if (r3 is not null) {
			result *= r3.Value;
		}

		return result;
	}

	//

	[Benchmark]
	public int Alg1Nullable_NotNull() {
		return Alg1Nullable(2, 3, 4, 5) ?? 0;
	}

	[Benchmark]
	public int Alg1Nullable_NotNull_NoInlining() {
		return Alg1Nullable_NoInlining(2, 3, 4, 5) ?? 0;
	}

	[Benchmark]
	public int Alg2Nullable_NotNull() {
		return Alg2Nullable(2, 3, 4, 5);
	}

	[Benchmark]
	public int Alg1Nullable_AllNull() {
		return Alg1Nullable(null, null, null, null) ?? 0;
	}

	[Benchmark]
	public int Alg2Nullable_AllNull() {
		return Alg2Nullable(null, null, null, null);
	}
}